from file_modules import compress2Tar
import sys, os

sys.path.insert(0, os.path.join(".."))
from cetl import build_pipeline


filepath = current_file_path = os.path.abspath(__file__)
current_dir = os.path.dirname(filepath)


c = compress2Tar(   target_dir=os.path.join(current_dir, "images"),
                    file_filter_regex="*.png",
                    out_dir=os.path.join(current_dir, "images"),
                    out_file = "images.tar.gz")

pipe = build_pipeline(c)

result = pipe.transform("")
print(result)


# python test_file_compress.py


# running command: python3.6 app.py
# -----------------------------------
# from sample_modules.add_new_column import addNewColumn
# from cetl import DataPipeline

# result=""
# cfg = {"pipeline":[
#                     {"type":"addRowNumber"},
#                     {"type":"parallelTransformer", "isParallel":1, "transformers":[
#                             {"type":"format2String", "description":"help following filterBy transformer"},
#                             {"pipeline":[   {"type":"filterBy"},
#                                             {"type":"parallelTransformer", "isParallel":1, "transformers":[
#                                                 {"type":"addRowNumber"},
#                                                 {"type":"passDataFrame"}]},
#                                             {"type":"toCSV", "out_dir":"./", "out_file":"out.csv"}]},
#                             {"pipeline":[{"type":"paddingZero"}]},
#                             {"type":"dropColumns", "description":"remove temp columns"}
#                         ]},
#                     {"type":"passDataFrame"},
#                     {"type":"addNewColumn", "base_field":"Tier", "value":"hi"},
#                     {"pipeline":[   {"type":"passDataFrame"},
#                                     {"type":"toCSV", "out_dir":"./", "out_file":"out.csv"}]}
#         ]}


# cfg = {"pipeline":[
#                     {"type":"generateDataFrame"},
#                     {"type":"addNewColumn", "base_field":"rownumber", "value":""},
#                     {"type":"addRowNumber", "mark_field":"rownumber"},
#                     {"type":"parallelTransformer", "transformers":[
                        

#                     ]}

# ]}



# from cetl import make_pipeline, DataPipeline
# from cetl.pandas_modules import generateDataFrame, unionAll
# from cetl.functional_modules import dummyStart, parallelTransformer, passDataFrame
# from cetl.utils.pipeline import _name_estimators



# print(_name_estimators([dummyStart(), passDataFrame()]))

    # pipe = make_pipeline(   dummyStart(),
    #                         parallelTransformer([generateDataFrame(), generateDataFrame()]), 
    #                         unionAll())
    # df = pipe.transform("")
    # print(df.to_markdown())
    # dot = pipe.build_digraph()
    # print(dot)

from cetl import build_pipeline
from cetl.pandas_modules import generateDataFrame, unionAll
from cetl.functional_modules import dummyStart, parallelTransformer

pipe = build_pipeline(   dummyStart(),
                        parallelTransformer([generateDataFrame(), generateDataFrame()]), 
                        unionAll())
df = pipe.transform("")
print(df)





# with DataPipeline(cfg) as pipe:
    # result = pipe.transform(result)
#     # print(pipe.steps)
#     # print(result)
#     # -----------------------------------------------
#     # create dot graph
#     pipe = pipe.build_digraph()
#     pipe.save_png("./sample5.png")

# ---------------------------------------------
# check all the module
# from cetl import PANDAS_TRANSFORMERS
# print(PANDAS_TRANSFORMERS.module_dict)




import hashlib
import datetime as date
import tarfile

class Block:
    def __init__(self, data, previous_hash):
        self.timestamp = date.datetime.now()
        self.data = data
        self.previous_hash = previous_hash
        self.hash = self.calculate_hash()

    def calculate_hash(self):
        sha = hashlib.sha256()
        sha.update(str(self.timestamp).encode('utf-8') +
                   str(self.data).encode('utf-8') +
                   str(self.previous_hash).encode('utf-8'))
        return sha.hexdigest()

class Blockchain:
    """
    # create blockchain
    my_blockchain = Blockchain()

    # read data from tar.gz file
    with tarfile.open("mydata.tar.gz", "r:gz") as tar:
        data = tar.extractfile("myfile.txt").read()

    # compute hash of data
    data_hash = hashlib.sha256(data).hexdigest()

    # create new block with hash and add to blockchain
    my_blockchain.add_block(Block(data_hash, ""))

    # print the blockchain
    my_blockchain.print_chain()
    """
    def __init__(self):
        self.chain = [self.create_genesis_block()]

    def create_genesis_block(self):
        return Block("Genesis Block", "0")

    def add_block(self, new_block):
        new_block.previous_hash = self.chain[-1].hash
        new_block.hash = new_block.calculate_hash()
        self.chain.append(new_block)

    def print_chain(self):
        for block in self.chain:
            print("Timestamp:", block.timestamp)
            print("Data:", block.data)
            print("Block Hash:", block.hash)
            print("Previous Hash:", block.previous_hash)
            print()


# pip install kafka-python
# pip install avro-python3
from kafka import KafkaProducer, KafkaConsumer, TopicPartition
from kafka.admin import KafkaAdminClient, NewTopic
import json
import pandas as pd
from kafka import KafkaProducer
import io
import pandas as pd
import pickle
from cryptography.fernet import Fernet
from kafka.errors import UnknownTopicOrPartitionError

class kafkaHelper:
    """usage

    """
    def __init__(self,
                bootstrap_servers=None,
                topic=None):

        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.producer=None
        self.consumer=None
        self.fernet_key = Fernet.generate_key()
        self.fernet = Fernet(self.fernet_key)
        self.admin_client = KafkaAdminClient(bootstrap_servers=self.bootstrap_servers)
        

    def __enter__(self,):
        
        try:
            # delete original topic
            self.delete_topic()

            # Create a Kafka producer instance
            self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

            # create instance of KafkaConsumer
            self.consumer = KafkaConsumer(self.topic,
                                    bootstrap_servers=self.bootstrap_servers,
                                    auto_offset_reset='earliest',
                                    enable_auto_commit=True)
        except Exception as e:
            if self.producer:
                self.producer.close()
            if self.consumer:
                self.consumer.close()
            print("has error")
        
        return self


    def __exit__(self, exc_type, exc_value, exc_tb):
        print("test when will run exit ################################")
        if self.producer:
            self.producer.close()
        if self.consumer:
            self.consumer.close()


    def topic_exists(self, admin_client, topic_name):
        """
        [{'error_code': 3, 'topic': 'c', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'o', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': '3', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'm', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 't', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'p', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'y', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'i', 'is_internal': False, 'partitions': []}]
        """
        try:
            topic_metadata = admin_client.describe_topics([topic_name])
            print(topic_metadata)
            # for partition_data in topic_metadata:
            #     if "error_code" in partition_data:

            # print("show topic_metadata", topic_metadata)
            if "error_code" in topic_metadata[0]:
                if topic_metadata[0]["error_code"]==3:
                    return False
                if topic_metadata[0]["error_code"]==0:
                    return True
            
            return True
        except UnknownTopicOrPartitionError:
            return False


    def delete_topic(self):   
        if self.topic_exists(self.admin_client, self.topic):
            print("delete topic ...")
            self.admin_client.delete_topics([self.topic])

    def send(self, key=None, value=None):
        if not isinstance(key, bytes):
            key = key.encode("utf-8")

        self.producer.send(self.topic, key=key, value=value)


    def receive(self, key):
        if not isinstance(key, bytes):
            key = key.encode("utf-8")
        for message in self.consumer:
            # print("message.key", message.key)
            # print("message.value", message.value)
            if message.key == key:
                received_message = message.value
                return received_message


    def close(self):
        self.producer.close()
        self.consumer.close()


    def serialize_dataframe(self, df):
        return pickle.dumps(df)

    
    def deserialize_dataframe(self, serialized_data):
        # deserialize the data
        file_obj = io.BytesIO(serialized_data)
        df = pickle.load(file_obj)
        return df


from .builder import FILE_TRANSFORMERS
import tarfile
from glob import glob
import os
import sys
sys.path.insert(0, os.path.join(".."))
from cetl import Base, transform_wrapper, context_name

@FILE_TRANSFORMERS.add()
class compress2Tar(Base):
    def __init__(   self, 
                    target_dir=None, 
                    file_filter_regex=None,
                    out_dir=None,
                    out_file=None):
        super().__init__()

        self.target_dir = target_dir
        self.file_filter_regex = file_filter_regex
        self.out_dir = out_dir
        self.out_file = out_file


    @transform_wrapper
    def transform(self, input):
        # retrieve filepaths
        filepaths = glob(os.path.join(self.target_dir, self.file_filter_regex))


        # compress file one by one into the out_archive
        out_archive = os.path.join(self.out_dir, self.out_file)
        with tarfile.open(out_archive, 'w:gz') as archive:
            for filepath in filepaths:
                archive.add(filepath)
            return {context_name:out_archive}
        



from sklearn.base import BaseEstimator, TransformerMixin

class Base(BaseEstimator, TransformerMixin):
    def __init__(self):
        
        self.exchange_media = "default" # e.g. "default", "kafka"
        self.bootstrap_servers=[]
        self.fernet_key=""
        self.broker_host = ""
        self.broker_port = ""
        # can be string or a list
        # if previous transformer is parallelTransformer, then would be a list
        self.pre_node_name = "" 
        self.delete_kafka_topic=1
        self.pipeline_start_time=""

        self.node_id = ""
        self.node_name = ""
        self.description = ""
        self.isParallel = "false"
        self.total_time = 0
        self.total_input = 0
        self.total_output = 0
        self.data_container_type="pandas"
        self.start_datetime = ""
        self.end_datetime = ""
        self.breakpoint = 0
        self.print_task_result=0
        self.is_first_trans = 0
        self.is_last_trans = 0

    def fit(self):
        pass
    


from pandas import DataFrame
from datetime import datetime

import sys
import os
sys.path.insert(0, os.path.join(".."))
from cetl import TEST_TRANSFORMERS as test_transformers
from cetl.utils.base import Base

@test_transformers.add()
class addNewColumn(Base):
    """
    {"type":"addNewColumn", "base_field":"Tier", "value":""}
    """
    def __init__(self, 
                base_field=None,
                value=None):

        self.base_field = base_field
        self.value = value

    def transform(self, dataframe: DataFrame):
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe
        df[self.base_field] = self.value
        
        return df


from .input import generateDataFrame
from .add_row_number import addRowNumber
from .add_new_column import (addNewColumn, addNewColumns, addNewColumnByPdSeries, addNewColumnFromExisting, 
                            addKeyColumnFromExisting, addUniqueKeyColumn, addNewFromConcatExistingColumns,
                            addNewFromMultiplyCalculation, addNewColumnWithNowDatetimeStr, 
                            addNewColumnWithDateAdjustment, addNewColumnWithDateTimeAdjustment, addNewColumnFromGradeHistory,
                            addNewColumnForIsWithinDayDiff, addNewColumnForDayDiff, addNewColumnForRowNumberToGroup,
                            addNewColumnFromEvalStatement, addGradeChangeType, addNewFromMinusCalculation,
                            addNewColumnWithReplaceDict)
from .filter import filterBy, filterValuesInList, filterByEvalStatement
from .pass_dataframe import passDataFrame, dummyStart, dummyStartEmpty
from .two2one import unionAll, sqlJoin


# run command: python test_kafka_helper
import pandas as pd

import sys, os
sys.path.insert(0, os.path.join(".."))
from cetl.utils.kafka_helper import kafkaHelper

# Define Kafka connection parameters
# althought it is not in docker container, 
# still need to use docker inspect to find the docker network
bootstrap_servers = ['172.31.0.1:9092']
topic_name = "mytopic"
# task_id
task_id = "1.filterBy"
# Create a pandas dataframe
df = pd.DataFrame({'col1': [1, 2, 3], 'col2': ['a', 'b', 'c']})

with kafkaHelper(   bootstrap_servers=bootstrap_servers,
                    topic=topic_name) as kh:

    # serialize the dataframe to pickle
    serialized_data = kh.serialize_dataframe(df)
    # print("serialized_data: ", serialized_data)

    # create instance of encryption
    encrypted_data = kh.fernet.encrypt(serialized_data)
    # print(encrypted_data)

    # send the data to kafka topic
    kh.send(key=task_id, value=encrypted_data)


    # received message
    message = kh.receive(task_id)

    # decrypt the message
    decrypted_data = kh.fernet.decrypt(message)

    # decode by pickle
    df = kh.deserialize_dataframe(decrypted_data)

    print(df)

    # delete the topic
    kh.delete_topic()


from .registry import Registry
import pandas as pd
DataFrame = pd.DataFrame
# pd.set_option('display.max_columns', None)

FUNCTIONAL_TRANSFORMERS = Registry("functional")
PANDAS_TRANSFORMERS = Registry("pandas")
JSON_TRANSFORMERS = Registry("json")
SPARK_TRANSFORMERS = Registry("spark")
TEST_TRANSFORMERS = Registry("test")
DB_MAPPERS = Registry("src_db_mappers")
DB_MODELS = Registry("db_models")
ALL_MODULES = {}
context_name="task_data"
default_module_type = "pandas"

def add_register(register=None):
    """usage
    from cetl.utils.register import Registry
    from cetl.utils.builder import add_register
    FILE_TRANSFORMERS = Registry("customer")
    add_register(register=FILE_TRANSFORMERS)
    """
    # get the register name
    register_name = register._name

    # check whether the register name already exist
    if register_name in ALL_MODULES:
        raise ValueError(f"The register name, {register_name} already exists")

    # add to ALL_MODULES
    ALL_MODULES[register_name] = register


for register in [   FUNCTIONAL_TRANSFORMERS,
                    PANDAS_TRANSFORMERS, 
                    JSON_TRANSFORMERS, 
                    SPARK_TRANSFORMERS, 
                    TEST_TRANSFORMERS,
                    DB_MAPPERS,
                    DB_MODELS]:

    add_register(register=register)



def get_register(module_type, default_type=None):
    # print("module_type:", module_type)
    TRANSFORMERS=None
    # module_type is like "pandas", "json" or "spark"
    # print("what is the content of all_modules", ALL_MODULES)

    if module_type=="":
        if not default_type:
            TRANSFORMERS = ALL_MODULES[default_module_type]
            return TRANSFORMERS
        else:
            TRANSFORMERS = ALL_MODULES[default_type]

    if module_type in ALL_MODULES:
        TRANSFORMERS = ALL_MODULES[module_type]
        return TRANSFORMERS
    elif default_type in ALL_MODULES:
        TRANSFORMERS = ALL_MODULES[default_type]
        return TRANSFORMERS
    else:
        print("ALL_MODULES", ALL_MODULES.keys())
        print(f"module_type, {module_type} is not recognized by cetl")
        assert False



def build_transformer_from_cfg(cfg, registry, parallel_transformers=None):
    args = cfg.copy()
    transformer_type = args.pop("type")
    
    if transformer_type not in registry.module_dict:
        print("current registry is ", registry._name)
        raise ValueError(f"transformer_type {transformer_type} not exists")

    cls_obj = registry.module_dict[transformer_type]
        
    
    transformer=None
    if transformer_type=="parallelTransformer":
        transformer = cls_obj(parallel_transformers)
        
    else:
        transformer = cls_obj(**args)

    return transformer


from functools import wraps
import time
from .builder import context_name, DataFrame
from datetime import datetime
from .file_mgt import get_datetime_str
from .kafka import kafka2python, python2kafka
EXCH_MEDIA = "default"


def recursive_records_count(count, data):
    if isinstance(data, dict):
        if context_name in data:
            count += len(data[context_name])
        elif "key" in data:
            return count
        else:
            print("not acceptable context_name")
            # raise ValueError()
            return count

    elif isinstance(data, DataFrame):
        count += data.shape[0]
    elif isinstance(data, str):
        count += 0
    elif isinstance(data, list):
        for item in data:
            count = recursive_records_count(count, item)
    elif isinstance(data, set):
        print("----------------------transformer output seems missing ':'", data)
        raise ValueError("please check the whether output should be dictionary since you are missing ':'")
    else:
        raise ValueError("currently only accept list, str, json and pandas data type")
    
    return count



def run_default_media(func, transformer, input, *args, **kwargs):

    # Calculate the execution time
    start_time = time.perf_counter()
    transformer.start_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")

    # print processing label
    print(f"Processing {transformer.node_name}, started at {transformer.start_datetime} ...") if transformer.print_task_result else ""

    # execute the transform function -----------------------------------------------------
    result = func(*args, **kwargs) #result here is module output

    # ----------------------------------------------------------------------------------------
    # Calcuate the end time
    end_time = time.perf_counter()
    total_time = end_time - start_time
    transformer.end_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
    # print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')


    # Adding the execution time to the transformer
    transformer.total_time = total_time
        
    # print(transformer)
    # parallelTransformer will make it run double times, not solved yet

    #count the number of records in input and output of the func:
    transformer.total_input = recursive_records_count(0, input)
    transformer.total_output = recursive_records_count(0, result)

    if transformer.node_name:
        print(f"    - take time {total_time:.4f} seconds, total input records {transformer.total_input}, total output records {transformer.total_output}") if transformer.print_task_result else ""

    # print(f"stop with breakpoint at {transformer.node_name}") if transformer.print_task_result else ""
    # print(f"{transformer.node_name}: take time {total_time:.4f} seconds, total input records {transformer.total_input}, total output records {transformer.total_output}")

    return result



def run_kafka_media(func, transformer, input, *args, **kwargs):

    if transformer.bootstrap_servers:

        from .kafka_media import kafkaMedia
        with kafkaMedia(bootstrap_servers=transformer.bootstrap_servers, 
                        topic=transformer.pipe_topic_name) as km:

            km.__enter__()

            # Calculate the execution time
            start_time = time.perf_counter()
            transformer.start_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")

            # print processing label
            print(f"Processing {transformer.node_name}, started at {transformer.start_datetime} ...") if transformer.print_task_result else ""

            # execute the transform function -----------------------------------------------------
            pre_module_output=""

            if transformer.is_first_trans:
                km.delete_topic()
                pre_module_output = ""

            else:

                if isinstance(input, dict):
                    # print(input)
                    pre_transformer_key = input["key"]
                    pre_module_output = km.read_kafka(task_id=pre_transformer_key)
                
                elif isinstance(input, list):
                    # previous transformer is wrapped by parallelTransformer
                    pre_module_output = []
                    for pre_transformer_key in [item["key"] for item in input]:
                        # ---------------------------------------------------------- read message
                        kafka2python = km.read_kafka(task_id=pre_transformer_key)
                        pre_module_output.append(kafka2python)

                elif isinstance(input, str):
                    # print("pre_transformer_key", pre_transformer_key)
                    # there is the in airflow, task relationship is not upstream
                    pre_transformer_key = input
                    pre_module_output = ""

            result = func(transformer, pre_module_output) # result here is module output

            # ----------------------------------------------------------------------------------------
            # Calcuate the end time
            end_time = time.perf_counter()
            total_time = end_time - start_time
            transformer.end_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
            # print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')


            # Adding the execution time to the transformer
            transformer.total_time = total_time
            
            # print(transformer)
            # parallelTransformer will make it run double times, not solved yet

            #count the number of records in input and output of the func:
            transformer.total_input = recursive_records_count(0, pre_module_output)
            transformer.total_output = recursive_records_count(0, result)

            if transformer.node_name:
                print(f"    - take time {total_time:.4f} seconds, total input records {transformer.total_input}, total output records {transformer.total_output}") if transformer.print_task_result else ""

                # print(f"stop with breakpoint at {transformer.node_name}") if transformer.print_task_result else ""
            # print(f"{transformer.node_name}: take time {total_time:.4f} seconds, total input records {transformer.total_input}, total output records {transformer.total_output}")

            # output result  -------------------------------------------------------------------------
            if transformer.exchange_media == "default":
                return result

            elif transformer.exchange_media =="kafka":
                # print(transformer)
                # print("####################")
                # sending the python object to kafka topic
                # print("transformer.node_name", transformer.node_name.encode('utf-8'))
                # print("result", result)
                km.send_kafka(task_id=transformer.node_name, value=result)

                result = {  "key":transformer.node_name,
                            "total_input":transformer.total_input, 
                            "total_output":transformer.total_output}
                return result
            else:

                print(transformer.exchange_media, transformer.node_name)
                assert False



def transform_wrapper(func):
    """
    Usage
    ----------------------------
    class Calculator:
        @transform_wrapper
        def calculate_something(transformer, num):
            total = sum((x for x in range(0, num**2)))
            return total

        def __repr__(transformer):
            return f'calc_object:{id(transformer)}'
    Reference
    ---------------------------
    https://dev.to/kcdchennai/python-decorator-to-measure-execution-time-54hk
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        transformer = args[0]
        input = args[1]
        if transformer.exchange_media=="default":
            return run_default_media(func, transformer, input, *args, **kwargs)
        else:
            return run_kafka_media(func, transformer, input, *args, **kwargs)

    return func_wrapper


    

    





import sys, os
sys.path.insert(0, os.path.join(".."))


from cetl.utils.registry import Registry
from cetl.utils.builder import add_register


FILE_TRANSFORMERS = Registry("file")
add_register(register=FILE_TRANSFORMERS)


# https://multimeric.github.io/PandasSchema/
import pandas as pd
from io import StringIO
from pandas_schema import Column, Schema
from pandas_schema.validation import LeadingWhitespaceValidation, TrailingWhitespaceValidation, CanConvertValidation, MatchesPatternValidation, InRangeValidation, InListValidation

schema = Schema([
    Column('Given Name', [LeadingWhitespaceValidation(), TrailingWhitespaceValidation()]),
    Column('Family Name', [LeadingWhitespaceValidation(), TrailingWhitespaceValidation()]),
    Column('Age', [InRangeValidation(0, 120)]),
    Column('Sex', [InListValidation(['Male', 'Female', 'Other'])]),
    Column('Customer ID', [MatchesPatternValidation(r'\d{4}[A-Z]{4}')])
])

test_data = pd.read_csv(StringIO('''Given Name,Family Name,Age,Sex,Customer ID
Gerald ,Hampton,82,Male,2582GABK
Yuuwa,Miyake,270,male,7951WVLW
Edyta,Majewska ,50,Female,775ANSID
'''))

errors = schema.validate(test_data)

for error in errors:
    print(error)


# pip install kafka-python
# pip install avro-python3
from kafka import KafkaProducer, KafkaConsumer, TopicPartition
from kafka.admin import KafkaAdminClient, NewTopic
from kafka.errors import UnknownTopicOrPartitionError
import json
import pandas as pd
from kafka import KafkaProducer
from avro import schema, io
import io
import pandas as pd
import pickle
from cryptography.fernet import Fernet



class kafkaMedia:
    def __init__(self,
                bootstrap_servers=None,
                topic=None,
                fernet_key=None):

        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.producer=None
        self.consumer=None
        self.fernet_key = fernet_key
        self.admin_client = KafkaAdminClient(bootstrap_servers=self.bootstrap_servers)


    def __enter__(self,):
        
        # self.delete_topic()

        # # Create a Kafka producer instance
        # self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

        # # create instance of KafkaConsumer
        # self.consumer = KafkaConsumer(self.topic,
        #                         bootstrap_servers=self.bootstrap_servers,
        #                         auto_offset_reset='earliest',
        #                         enable_auto_commit=True)

        try:
            # self.admin_client = KafkaAdminClient(bootstrap_servers=self.bootstrap_servers)
            # self.delete_topic()

            # Create a Kafka producer instance
            self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

            self.consumer = KafkaConsumer(self.topic,
                                    bootstrap_servers=self.bootstrap_servers,
                                    auto_offset_reset='earliest',
                                    enable_auto_commit=True)

        except Exception as e:
            if self.producer:
                self.producer.close()
            if self.consumer:
                self.consumer.close()
            print("has error")
        
        return self


    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def topic_exists(self, admin_client, topic_name):
        try:
            topic_metadata = admin_client.describe_topics([topic_name])
            # for partition_data in topic_metadata:
            #     if "error_code" in partition_data:

            # print("show topic_metadata", topic_metadata)
            if "error_code" in topic_metadata[0]:
                if topic_metadata[0]["error_code"]==3:
                    return False

                if topic_metadata[0]["error_code"]==0:
                    return False
            
            return True
        except UnknownTopicOrPartitionError:
            return False


    def delete_topic(self):
        # if not self.admin_client:
            # self.__enter__()
        if self.topic_exists(self.admin_client, self.topic):
            print("delete kafka topic ...")
            self.admin_client.delete_topics([self.topic])


    def send(self, key=None, value=None):
        # if not self.producer:
            # self.__enter__()
        # print("topic", self.topic)
        # print("send key", key)
        if not isinstance(key, bytes):
            key = key.encode("utf-8")

        try:
            self.producer.send(self.topic, key=key, value=value)
        except Exception as e:
            print(self.producer)
            print("key", key)
            print("value", value)
            print(e)
        # producer.flush()

    def send_kafka(self, task_id=None, value=None):


        if not isinstance(task_id, bytes):
            task_id = task_id.encode('utf-8')

        df = value

        # serialize the dataframe to pickle
        serialized_data = self.serialize_dataframe(df)
        # print("serialized_data: ", serialized_data)

        # create instance of encryption
        encrypted_data=""
        if self.fernet_key:
            fernet = Fernet(self.fernet_key)
            encrypted_data = fernet.fernet.encrypt(serialized_data)
            self.send(key=task_id, value=encrypted_data)
        else:
            self.send(key=task_id, value=serialized_data)

        # send the data to kafka topic
        # print("send key", task_id)
        
        # print("send encrypted_data", encrypted_data)


    def receive(self, key=None):
        # if not self.consumer:
        #     # create instance of KafkaConsumer
        #     self.consumer = KafkaConsumer(self.topic,
        #                             bootstrap_servers=self.bootstrap_servers,
        #                             auto_offset_reset='earliest',
        #                             enable_auto_commit=True)

        # print("self.consumer", self.consumer)
        for message in self.consumer:
            # print("dectect message", message.key, key)
            # print(message.key, key)
            if message.key == key:
                received_message = message.value
                return received_message


    def receive_with_partition(self, key=None):
        # if not self.consumer:
        #     # create instance of KafkaConsumer
        #     self.consumer = KafkaConsumer(self.topic,
        #                             bootstrap_servers=self.bootstrap_servers,
        #                             auto_offset_reset='earliest',
        #                             enable_auto_commit=True)

        # Get partitions for the topic
        partitions = self.consumer.partitions_for_topic(self.topic)

        # Loop through partitions and clear messages
        if partitions:
            for partition in partitions:
                # Get current partition offset
                tp = TopicPartition(self.topic, partition)
                current_offset = self.consumer.position(tp)

                # Seek to beginning of partition
                self.consumer.seek_to_beginning(tp)

                # Consume messages and send null messages with same keys to clear the topic
                for message in self.consumer:
                    # print("detect message", message.key, key)
                    if message.key == key:
                        received_message = message.value
                        return received_message
        else:
            return self.receive(key=key)
            
            

    def read_kafka(self, task_id=None):
        # received message
        if not isinstance(task_id, bytes):
            task_id = task_id.encode('utf-8')

        message = self.receive_with_partition(key=task_id)
        assert message
        
        if self.fernet_key:
            # decrypt the message
            fernet = Fernet(self.fernet_key)
            decrypted_data = fernet.decrypt(message)
            return self.deserialize_dataframe(decrypted_data)
        else:
            return self.deserialize_dataframe(message)


    def close(self):
        if self.producer:
            self.producer.close()
        if self.consumer:
            self.consumer.close()


    def serialize_dataframe(self, df):
        return pickle.dumps(df)

    
    def deserialize_dataframe(self, serialized_data):
        # deserialize the data
        file_obj = io.BytesIO(serialized_data)
        df = pickle.load(file_obj)
        return df


# from .kafka_helper import kafkaHelper
# from ..utils.builder import pd
# from cryptography.fernet import Fernet

# class kafkaMedia(kafkaHelper):
#     def __init__(self, 
#                 bootstrap_servers=None,
#                 topic=None):
#         super().__init__(bootstrap_servers=bootstrap_servers,
#                         topic=topic)

#     def send_kafka(self, key=None, value=None):

#         if isinstance(value, pd.DataFrame):
#             df = value

#             # serialize the dataframe to pickle
#             serialized_data = self.serialize_dataframe(df)
#             # print("serialized_data: ", serialized_data)

#             # create instance of encryption
#             self.pub_key = Fernet.generate_key()
#             self.fernet = Fernet(self.pub_key)

#             encrypted_data = self.fernet.encrypt(serialized_data)
#             # print("encrypted_data", encrypted_data)

#             # send the data to kafka topic
#             self.send(key=key, value=encrypted_data)

#     def read_kafka(self, key=None):

        
#         # received message
#         message = self.receive(key=key)

#         # decrypt the message
#         decrypted_data = self.fernet.decrypt(message)

#         # decode by pickle
#         df = self.deserialize_dataframe(decrypted_data)

#         # delete the topic
#         # kh.delete_topic()

        return df


